#include "Arduino.h"

void mtp_reset(void);
void mtp_loop(void);

void blink(uint32_t del)
{ static uint32_t to;
  if(millis()-to > del)
  {
    digitalWriteFast(LED_BUILTIN,!digitalReadFast(LED_BUILTIN));
    to=millis();
  }
}

void log(uint32_t del, const char *txt)
{ static uint32_t to;
  if(millis()-to > del)
  {
    Serial.println(txt);
    to=millis();
  }
}

void setup()
{ 
  while(!Serial);
  Serial.println("MTP test");  Serial.flush();
  pinMode(LED_BUILTIN,OUTPUT);
  digitalWriteFast(LED_BUILTIN,HIGH);
  mtp_reset();
  digitalWriteFast(LED_BUILTIN,LOW);
}

void loop()
{ blink(500);

  mtp_loop();

  log(1000,"loop");
}


#include "usb_mtp.h"
#include "debug/printf.h"

extern "C"  uint32_t mtp_tx_counter;
extern "C"  uint32_t mtp_rx_counter;
extern "C"  uint32_t mtp_rx_event_counter;
extern "C"  uint32_t mtp_tx_event_counter;


// Container Types
#define MTP_CONTAINER_TYPE_UNDEFINED    0
#define MTP_CONTAINER_TYPE_COMMAND      1
#define MTP_CONTAINER_TYPE_DATA         2
#define MTP_CONTAINER_TYPE_RESPONSE     3
#define MTP_CONTAINER_TYPE_EVENT        4

// Container Offsets
#define MTP_CONTAINER_LENGTH_OFFSET             0
#define MTP_CONTAINER_TYPE_OFFSET               4
#define MTP_CONTAINER_CODE_OFFSET               6
#define MTP_CONTAINER_TRANSACTION_ID_OFFSET     8
#define MTP_CONTAINER_PARAMETER_OFFSET          12
#define MTP_CONTAINER_HEADER_SIZE               12

// MTP Operation Codes
#define MTP_OPERATION_GET_DEVICE_INFO                       0x1001
#define MTP_OPERATION_OPEN_SESSION                          0x1002
#define MTP_OPERATION_CLOSE_SESSION                         0x1003
#define MTP_OPERATION_GET_STORAGE_IDS                       0x1004
#define MTP_OPERATION_GET_STORAGE_INFO                      0x1005
#define MTP_OPERATION_GET_NUM_OBJECTS                       0x1006
#define MTP_OPERATION_GET_OBJECT_HANDLES                    0x1007
#define MTP_OPERATION_GET_OBJECT_INFO                       0x1008
#define MTP_OPERATION_GET_OBJECT                            0x1009
#define MTP_OPERATION_GET_THUMB                             0x100A
#define MTP_OPERATION_DELETE_OBJECT                         0x100B
#define MTP_OPERATION_SEND_OBJECT_INFO                      0x100C
#define MTP_OPERATION_SEND_OBJECT                           0x100D
#define MTP_OPERATION_INITIATE_CAPTURE                      0x100E
#define MTP_OPERATION_FORMAT_STORE                          0x100F
#define MTP_OPERATION_RESET_DEVICE                          0x1010
#define MTP_OPERATION_SELF_TEST                             0x1011
#define MTP_OPERATION_SET_OBJECT_PROTECTION                 0x1012
#define MTP_OPERATION_POWER_DOWN                            0x1013
#define MTP_OPERATION_GET_DEVICE_PROP_DESC                  0x1014
#define MTP_OPERATION_GET_DEVICE_PROP_VALUE                 0x1015
#define MTP_OPERATION_SET_DEVICE_PROP_VALUE                 0x1016
#define MTP_OPERATION_RESET_DEVICE_PROP_VALUE               0x1017
#define MTP_OPERATION_TERMINATE_OPEN_CAPTURE                0x1018
#define MTP_OPERATION_MOVE_OBJECT                           0x1019
#define MTP_OPERATION_COPY_OBJECT                           0x101A
#define MTP_OPERATION_GET_PARTIAL_OBJECT                    0x101B
#define MTP_OPERATION_INITIATE_OPEN_CAPTURE                 0x101C
#define MTP_OPERATION_GET_OBJECT_PROPS_SUPPORTED            0x9801
#define MTP_OPERATION_GET_OBJECT_PROP_DESC                  0x9802
#define MTP_OPERATION_GET_OBJECT_PROP_VALUE                 0x9803
#define MTP_OPERATION_SET_OBJECT_PROP_VALUE                 0x9804
#define MTP_OPERATION_GET_OBJECT_PROP_LIST                  0x9805
#define MTP_OPERATION_SET_OBJECT_PROP_LIST                  0x9806
#define MTP_OPERATION_GET_INTERDEPENDENT_PROP_DESC          0x9807
#define MTP_OPERATION_SEND_OBJECT_PROP_LIST                 0x9808
#define MTP_OPERATION_GET_OBJECT_REFERENCES                 0x9810
#define MTP_OPERATION_SET_OBJECT_REFERENCES                 0x9811
#define MTP_OPERATION_SKIP                                  0x9820

uint8_t data_buffer[MTP_RX_SIZE] __attribute__ ((used, aligned(32)));

  struct MTPHeader {
    uint32_t len;  // 0
    uint16_t type; // 4
    uint16_t op;   // 6
    uint32_t transaction_id; // 8
  };

  struct MTPContainer {
    uint32_t len;  // 0
    uint16_t type; // 4
    uint16_t op;   // 6
    uint32_t transaction_id; // 8
    uint32_t params[5];    // 12
  } __attribute__((__may_alias__))  ;

#define CONTAINER ((struct MTPContainer*)(data_buffer))

void mtp_event(int code, int session)
{ struct MTPHeader event;
  event.len=sizeof(event);
  event.type= MTP_CONTAINER_TYPE_EVENT;
  event.op=code;
  event.transaction_id=0;
  usb_mtp_eventSend((void *) &event,30);
}

void mtp_reset(void)
{
  mtp_event(0x4008,1);
}

void mtp_loop(void)
{
  if(mtp_rx_counter>0)
  { mtp_rx_counter=0;

    usb_mtp_recv(data_buffer,30);
    printf("RX: %d %d %x %d %d\n", 
                CONTAINER->len,CONTAINER->type,CONTAINER->op, 
                CONTAINER->transaction_id,CONTAINER->params[0]);

    switch (CONTAINER->op)
    {
      case 0x1002:  //open session
        int p1=CONTAINER->params[0];
        CONTAINER->type=3;
        CONTAINER->op=0x2001; // OK
        CONTAINER->params[0]=p1;
        for(int ii=1; ii<5;ii++)  CONTAINER->params[ii]=0;
        printf("TX: %d %d %x %d %d\n", 
                    CONTAINER->len,CONTAINER->type,CONTAINER->op, 
                    CONTAINER->transaction_id,CONTAINER->params[0]);
        usb_mtp_send(data_buffer,30);
        break;

      default:
        while(1) blink(100); // lock Teensy
    }
  }

  if(mtp_tx_counter>0)
  {  
    mtp_tx_counter=0;
  }

  if(mtp_rx_event_counter>0)
  {
    mtp_rx_event_counter=0;
  }

  if(mtp_tx_event_counter>0)
  {
    mtp_tx_event_counter=0;
  }

  { 
    static uint32_t to=0;
    if(millis()-to > 5000)
    {
      mtp_reset();
      to=millis();
    }
  }
}

#if 0 // final mtp_responder code
#define MTP  1
  #if MTP==1
    #include "mtp_responder.h"

    MTPStorage storage;
    MTPD       mtpd(&storage);
  #endif

  #if MTP==1
    mtpd.loop();
  #endif
#endif